package edu.georgetown.gucs.matcher;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Serializable;
import java.util.Comparator;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.SortedMap;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.Random;
import java.math.BigInteger;


import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.fingerprinter.BulkFingerprinter;
import edu.georgetown.gucs.matcher.BulkMatcher;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.ObjectBuilder;
import edu.georgetown.gucs.utility.Pair;
import edu.georgetown.gucs.utility.Token;
import edu.georgetown.gucs.utility.BoundedExecutor;



/**
 * Outputs a list of files and their scores, comparing a given fingerprint to a directory of files. Files are organized
 * by scores with higher scores appearing higher in the list. Providing a minimum score and a dictionary (if the
 * fingerprint does not already contain one) is optional.
 * 
 * @author Lindsay Neubauer
 */
public class CompareFingerprint implements Iterable<Entry<Integer, List<String>>> {

  private Fingerprinter fingerprint;
  private FingerprintMatcher matcher;
  private FileLister fileLister;
  private int minimum_score;
  private Dictionary dictionary;
  private SortedMap<Integer, List<String>> scores;
  private int maxThread;

  /** Comparator that orders scores highest to lowest */
  private static class HightoLowScore implements Comparator<Integer>, Serializable {
    private static final long serialVersionUID = 5679507537454070748L;

    public HightoLowScore() {
    }

    @Override
    public int compare(Integer one, Integer two) {
      return two.compareTo(one);
    }
  }

  /**
   * Constructor that sets the matcher, fingerprint file, directory, minimum score and dictionary to use for this
   * comparison
   * 
   * @param match
   *          the string name of the matcher to use for this comparison
   * @param fingerprintFile
   *          the string filename of the fingerprint to compare with this directory
   * @param directory
   *          the string directory path to compare with this fingerprint
   * @param minimumScore
   *          the <code>int</code> minimum score needed to output a file for this comparison
   * @param dictionary
   *          the string filename of the dictionary to use for this comparison; this dictionary will only be used if the
   *          fingerprint file does not already contain a full dictionary
   */
  public CompareFingerprint(String match, String fingerprintFile, String directory, int minimumScore, String dictionary) {
      this.matcher = (FingerprintMatcher) ObjectBuilder.buildObject(match);
    this.scores = Collections.synchronizedSortedMap(new TreeMap<Integer, List<String>>(new HightoLowScore()));
    this.fileLister = new FileLister(directory);
    
    this.minimum_score = (minimumScore < 0 || minimumScore > 99) ? 0 : minimumScore;

    this.fingerprint = new Fingerprinter(fingerprintFile, dictionary);
    this.dictionary = this.fingerprint.getDictionary();
    this.maxThread = Runtime.getRuntime().availableProcessors();
  }

  @Override
  public Iterator<Entry<Integer, List<String>>> iterator() {
    return this.scores.entrySet().iterator();
  }

  /**
   * Compute a fingerprint for each file in this directory and compare it to this fingerprint using this matcher; if the
   * matcher returns boolean values, true values are given a score of 99 and false values are given a score of 0
   */
  public void compareFingerprint() {
    Global.monitor.writeToBuffer("CompareFingerprint", "compareFingerprint",
        "directory=" + this.fileLister.getDirectory());
    Global.monitor.printMessage("CompareFingerprint", "Comparing fingerprint");

    Fingerprinter fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(this.fingerprint.getFingerprintName());
    fingerprinter.setDictionary(this.dictionary);
    String matcherName = this.matcher.getClass().getSimpleName();
    int score;

    for (File file : this.fileLister) {
      for (Pair<String, Token> fingerprintPair : fingerprinter.computeFingerprintXML(file.getPath())) {
        for (Pair<String, Token> thisFingerprint : this.fingerprint.getBase64Fingerprints()) {
          if (matcherName.equals("HashFingerprintMatcher") || matcherName.equals("ExactFingerprintMatcher")) {
            score = (this.matcher.match(thisFingerprint.getA(), fingerprintPair.getA())) ? 99 : 0;
          } else {
            score = this.matcher.getScoreXML(thisFingerprint.getA(), fingerprintPair.getA());
          }

          List<String> updatedScores;
          if (this.scores.containsKey(score)) {
            updatedScores = this.scores.get(score);
          } else {
            updatedScores = new Vector<String>();
          }
          updatedScores.add(fingerprintPair.getB().toString());
          this.scores.put(score, updatedScores);
        }
      }
    }
  }

  /**
   * Compute a fingerprint for each file in this directory and compare it to each other using this matcher; if the
   * matcher returns boolean values, true values are given a score of 99 and false values are given a score of 0
   */
  public void compareFingerprintDirectory() {

      Global.monitor.writeToBuffer("Compare","compareFingerprintDirectory","Starting");

      List<Pair<byte[], Token>> results;

      
      
      BulkFingerprinter bulkFP = new BulkFingerprinter(this.fingerprint.getFingerprintName(),this.fileLister,this.dictionary, maxThread);

    Global.monitor.writeToBuffer("CompareFingerprint", "compareFingerprintDirectory",
        "directory=" + this.fileLister.getDirectory());
    Global.monitor.writeToBuffer("Compare","compareFingerprintDirectory","Calling BulkFingerprinter");
    
    Global.monitor.writeToBuffer("Compare","compareFingerprintDirectory","Fingerprinting");
    
    System.err.println("Fingerprinting" + System.currentTimeMillis());

    results = bulkFP.computeAllFingerprints(this.fingerprint.getFingerprintName(),this.dictionary,this.fileLister);
    
    String matcherName = this.matcher.getClass().getSimpleName();

    Global.monitor.writeToBuffer("Compare","compareFingerprintDirectory","Calling BulkMatcher");

    Global.monitor.writeToBuffer("Compare","compareFingerprintDirectory","Matching");

    System.err.println("Matching" + System.currentTimeMillis());
    
    BulkMatcher bulkMatch = new BulkMatcher(matcherName,scores,0,maxThread);
    bulkMatch.compareAllFingerprints(results);
    
    Global.monitor.writeToBuffer("Compare","compareFingerprintDirectory","Complete");

  }

  /**
   * Provides all the files and their similarity scores; if a minimum score is provided, only output those files with
   * scores greater than or equal to the minimum
   * 
   * @return the string containing all the score results
   */
  private String getScores() {
    StringBuffer buffer = new StringBuffer();
    boolean printedSomething = false;
    for (int i = 99; i >= this.minimum_score; i--) {
      if (this.scores.containsKey(i)) {
        for (String score : this.scores.get(i)) {
          buffer.append(String.format("%s%n%n", score + ": Score = " + i));
        }
        buffer.append(String.format("%n---------------------->%n"));
        printedSomething = true;
      }
    }
    if (!printedSomething) {
      buffer.append(String.format("%s%n",
          "CompareDictionary: There are no files in the directory with a score higher than " + this.minimum_score));
    }
    return buffer.toString();
  }

  /**
   * Print all the files and their similarity scores; if a minimum score is provided, only output those files with
   * scores greater than or equal to the minimum
   */
  public void printScores() {
    System.out.println(getScores());
  }

  /**
   * Write all the files and their similarity scores to a file; if a minimum score is provided, only output those files
   * with scores greater than or equal to the minimum
   * 
   * @param filename
   *          the string name of the file to write output
   */
  public void writeScores(String filename) {
    Global.monitor.writeToBuffer("CompareFingerprint", "writeScores", "filename=" + filename);
    try {
      OutputStreamWriter out = new OutputStreamWriter(new FileOutputStream(filename), "UTF-8");
      BufferedWriter writer = new BufferedWriter(out);
      writer.write(getScores());
      writer.close();
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    }
    System.out.println("Writing scores to " + filename);
  }

  /**
   * Outputs a list of files and their scores, comparing a given fingerprint to a directory of files. Files are
   * organized by scores with higher scores appearing higher in the list. Providing a minimum score and a dictionary (if
   * the fingerprint does not already contain one) is optional.
   * 
   * @param args
   *          array of string command line arguments<br>
   *          <code>args[0]</code> the name of the matcher to use for this comparison<br>
   *          <code>args[1]</code> the filename of the fingerprint to compare with this directory<br>
   *          <code>args[2]</code> the directory path to compare with this fingerprint<br>
   *          <code>args[3]</code> the optional integer minimum score needed to output a file for this comparison<br>
   *          <code>args[4]</code> the optional boolean indicating true if n-squared comparison desired, default: false<br>
   *          <code>args[5]</code> the optional filename of the dictionary to use for this comparison; only used if the
   *          fingerprint file does not contain a full dictionary or a link to a dictionary in the file system
   */
  public static void main(String args[]) {
    try {
      String matcher = args[0];
      String fingerprint = args[1];
      String directory = args[2];
      int min_score = (args.length > 3) ? Integer.valueOf(args[3]) : 0;
      boolean n_squared = (args.length > 4) ? Boolean.valueOf(args[4]) : false;
      String dictionary = (args.length > 5) ? args[5] : null;

      CompareFingerprint compare = new CompareFingerprint(matcher, fingerprint, directory, min_score, dictionary);
      if (!n_squared) {
        compare.compareFingerprint();
      } else {
        compare.compareFingerprintDirectory();
      }
      compare.printScores();
      Global.monitor.closeBuffer();
    } catch (ArrayIndexOutOfBoundsException e) {
      System.err
          .println("Usage: java CompareFingerprint <matcher name> <XML fingeprint file> <directory> <optional: minimum score> "
              + "<optional: true/false for n-squared directory comparison> <optional: dictionaryFilename>");
      System.err
          .println("Outputs a list of files and their scores, comparing a given fingerprint to a directory of files");
      System.exit(-1);
    } catch (RuntimeException exception) {
      ExceptionHandler.handleExceptionExit(exception);
    }
  }
}
