package edu.georgetown.gucs.tokenizers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Collections;
import java.util.Dictionary;
import java.util.List;
import java.util.Vector;
import java.util.zip.GZIPInputStream;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Token;

/**
 * Splits contents of a gzipped text file into tokens based on whitespace or by line. Used for {@linkplain Dictionary}
 * creation.
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class GzippedFileTokenizer extends FileTokenizer {
  private static final long serialVersionUID = -1100868800293796306L;

  /**
   * Constructor that sets the token creation mode to split based on whitespace
   */
  public GzippedFileTokenizer() {
    this.mode = "tokens";
  }

  /**
   * Constructor that sets the token creation mode. If set to "lines" the file is split by line; if set to "tokens" the
   * file is split based on whitespace.
   * 
   * @param mode
   *          the string mode to split tokens based on whitespace or by line
   */
  public GzippedFileTokenizer(String mode) {
    this.mode = mode.trim();
  }

  /**
   * Splits a gzipped text document into tokens
   */
  @Override
  public List<Token> tokenize(String filename) {
    if (!filename.contains(".gz")) {
      Global.monitor.printErrorMessage("GzippedFileTokenizer", "Given filename (" + filename
          + ") does not contain gzip file extension");
      Global.monitor.writeToBuffer("GzippedFileTokenizer", "tokenize",
          "Given filename does not contain gzip file extension");
      ExceptionHandler.handleExceptionThrow(new Exception("GzippedFileTokenizer: Given filename (" + filename
          + ") does not contain gzip file extension"));
      return null;
    }
    return readFile(filename);
  }

  /**
   * Splits a gzipped text document into tokens.
   */
  @Override
  protected List<Token> readFile(String filename) {
    BufferedReader buffer_reader = null;
    try {
      // TODO GzippedFileTokenizer: Assumes file is UTF-8 encoded (otherwise, it will spit out garbage)
      buffer_reader = new BufferedReader(new InputStreamReader(new GZIPInputStream(new FileInputStream(filename)),
          "UTF-8"));

      StringBuffer token = new StringBuffer("");
      int character;
      int byte_start = 0;
      int byte_end = 0;

      synchronized (this.tokenVectorMap) {
        this.tokenVectorMap.clear();
        List<Token> tokenVector = Collections.synchronizedList(new Vector<Token>());
        synchronized (tokenVector) {
          while ((character = buffer_reader.read()) != -1) {
            if (this.mode.equalsIgnoreCase("lines")) {
              if ('\n' == character || '\r' == character || '\f' == character) {
                if (token.toString().trim().length() > 0) {
                  tokenVector.add(new Token(token.toString().trim(), byte_start, byte_end));
                  byte_start = byte_end;
                }
                byte_end += Character.charCount(character);
                token.delete(0, token.length());
              } else {
                token.append((char) character);
                byte_end += Character.charCount(character);
              }
            } else {
              if (Character.isWhitespace(character)) {
                if (token.toString().trim().length() > 0) {
                  tokenVector.add(new Token(token.toString().trim(), byte_start, byte_end));
                  byte_start = byte_end;
                }
                byte_end += Character.charCount(character);
                token.delete(0, token.length());
              } else {
                token.append((char) character);
                byte_end += Character.charCount(character);
              }
            }
          }
          buffer_reader.close();
          if (token.toString().trim().length() > 0) {
            tokenVector.add(new Token(token.toString().trim(), byte_start, byte_end));
          }
          this.tokenVectorMap.put("null", tokenVector);
          return tokenVector;
        }
      }
    } catch (IOException e) {
      ExceptionHandler.handleExceptionExit(e);
    } finally {
      try {
        if (buffer_reader != null) {
          buffer_reader.close();
        }
      } catch (IOException e) {
        ExceptionHandler.handleExceptionExit(e);
      }
    }
    return null;
  }
}