package edu.georgetown.gucs.tokenizers;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Vector;

import edu.georgetown.gucs.utility.Token;

/**
 * Eliminates tokens specified in the given stop words document; Assumes this document contains lower case,
 * non-porterized English word
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 * 
 */
public class StopWordRemoverTokenizer extends Tokenizer {

  private static final long serialVersionUID = -1767358641887650726L;
  private Map<String, Object> stopwords = new HashMap<String, Object>();

  /**
   * Constructor that loads stop words to use when <code>tokenize</code> method is called;
   * 
   * @param stopwordFilename
   *          the string filename of the document containing the stop words to use; assumes the document contains lower
   *          case, non-porterized English word
   */
  public StopWordRemoverTokenizer(String stopwordFilename) {
    this.stopwords = new HashMap<String, Object>();
    loadStopWords(stopwordFilename);
  }

  /**
   * Loads words to eliminate when <code>tokenize</code> method is called
   * 
   * @param stopwordFilename
   *          the string filename of the document containing the stop words to use; assumes the document contains lower
   *          case, non-porterized English word
   */
  public void loadStopWords(String stopwordFilename) {
    FileTokenizer fileTokenizer = new FileTokenizer();
    fileTokenizer.tokenize(stopwordFilename);
    Iterator<Token> iterator = fileTokenizer.getTokenVectorMap().get("null").iterator();
    while (iterator.hasNext()) {
      this.stopwords.put(iterator.next().name(), null);
    }
  }

  /**
   * Eliminates tokens specified in the given stop words document
   */
  @Override
  public List<Token> tokenize(List<Token> tokens) {
    List<Token> tokenVector = Collections.synchronizedList(new Vector<Token>());
    synchronized (tokenVector) {
      for (Token token : tokens) {
        String lctoken = token.name().toLowerCase(Locale.ENGLISH);
        if (!this.stopwords.containsKey(lctoken)) {
          tokenVector.add(token);
        }
      }
      return tokenVector;
    }
  }
}
