package edu.georgetown.gucs.utility;

import edu.georgetown.gucs.utility.Pair;

import java.io.File;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Queue;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.Vector;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.Random;
import java.math.BigInteger;


/**
 * Creates a list of files in a directory and allows iteration over all and parts of this list
 * 
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class FileLister implements Iterable<File> {

    private File directory;
    // private ExecutorService executor;
    //  private BoundedExecutor boundExecutor;
    
    // This is a blocking queue that can be drawn from while the FileLister is still reading
    // the directory structure; useful for threaded dictionary creation
    private LinkedBlockingQueue<File> fileQueue = new LinkedBlockingQueue<File>();

    // This is the list of all files found. 
    private List<File> fileList = new ArrayList<File>();

    // These are used for experiments that require a sample of files for 
    // Dicitonary creation and a separate sample for testing
    // This is used to create a subset of files for dictionary creation  
    private List<File> dictionaryList = new ArrayList<File>();
    // This is used in combination with the dicitonary list to provide a set of files that were
    // not used in Dictionary creation
    private List<File> sampleList = new ArrayList<File>();


    private boolean loaded = false;
    private int maxThread;
    private BlockingQueue<File> sharedDirectoryQueue = new LinkedBlockingQueue<File>();

	 
     /**
     * Constructor that recursively loads files from the given directory and sets the maximum thread count to the number
     * of processors available to the Java virtual machine
     * 
     * @param directory
     *          the string path to the directory to load
     */
    public FileLister(String directory) {
	this.directory = new File(directory);
	maxThread = (Runtime.getRuntime().availableProcessors());
	
	recursiveLoadDirectory(directory);    
    }

    /**
     * Constructor that recursively loads files from the given directory and sets the maximum thread count
     * 
     * @param directory
     *          the string path to the directory to load
     * @param maxThread
     *          the <code>int</code> maximum threads to use to run this fileLister
     */
    public FileLister(String directory, int maxThread) {
	this.directory = new File(directory);
	this.maxThread = maxThread;
	recursiveLoadDirectory(directory);
    }


    /**
     * Provides an iterator over the list of all files
     * 
     * @return the file iterator over the list of all files
     */
    @Override
	public Iterator<File> iterator() {
	return this.fileList.iterator();
    }

    /**
     * Provides the size of the list of all files
     * 
     * @return the <code>int</code> number of files
     */
    public int size() {
	return this.fileList.size();
    }

   

    /**
     * Provides the boolean indicating if all files have been loaded into the linked blocking queue
     * 
     * @return the boolean value indicating if all files have been loaded
     */
    public boolean getLoaded() {
	return this.loaded;
    }

    /**
     * Lets a process that is taking files off the linked blocking queue know if there are more
     * files or not; use to test if there will be any more.
     */
    public boolean moreFiles(){

	// Done means if we are done checking, not if we are done with files
	boolean done = false;
	// More is if there are more files or not
	boolean more = false;

	//Global.monitor.writeToBuffer("FileLister", "moreFiles", "Entering");
	

	while (!done){
	    try {
		if (this.fileQueue.peek() != null) {
		    done = true;
		    more = true;
		}
		if ((this.loaded == true) && (this.fileQueue.peek() == null) ){
		    done = true;
		    more = false;
		}
		
		if (!done) {		    
		    // We just don't know yet. Wait a sec and I'll check
		    Thread.sleep(100);						
		}
		
	    } catch (InterruptedException e){
		ExceptionHandler.handleExceptionExit(e);
	    }
	    
	}

	//Global.monitor.writeToBuffer("FileLister", "moreFiles", "Exiting with: " + more);
	return more;

    }


    /**
     * Sets the boolean indicating if all files have been loaded to true
     * Only the FileListerHR thread should call this
     */
    void setLoaded() {
	//System.err.println("Done loading, have " + fileList.size() + " files");
	this.loaded = true;
    }


    /**
     * Provides the linked blocking queue of all the files in the directory
     * 
     * @return the linked blocking queue of all the files in the directory
     */
    public LinkedBlockingQueue<File> getFileQueue() {
	return this.fileQueue;
    }

    /**
     * Provides a list of all files in the directory
     * 
     * @return the list of all the files in the directory
     */
    public List<File> getFileList() {
	return this.fileList;
    }

    /**
     * Determines whether all the files have been loaded and unloaded from the queue of files
     * 
     * @return the <code>boolean</code> value whether all the files have been loaded and unloaded from the queue of files
     */
    public synchronized boolean loadComplete() {
	    return (this.loaded);
	}

    /**
     * Provides the path to the directory used for the initial list of files
     * 
     * @return the string path to the directory used to initially create this list of files
     */
    public File getDirectory() {
	return this.directory;
    }

    /**
     * Gets the next directory to be processed by a thread. Increments the active thread count so we know
     * when all the threads are done
     * 
     **/

    File getNextDirectory() {
	    
	File nextDir = null;
	try {
	    nextDir = this.sharedDirectoryQueue.take();
	}
	catch (InterruptedException e){
	    ExceptionHandler.handleExceptionExit(e);
	}

	return nextDir;
    }
	
    /**
     * Decrements the active thread count when a thread finishes processing a directory
     *
     **/

    /**
     * Add a new directory to the list of directories to be processed.
     *
     **/
    void addDirectory(File newDir){
	    
	    try {
		Global.monitor.writeToBuffer("FileLister", "addDirectory", "Dic is: " + newDir + " Thread is " + Thread.currentThread().getName());
		sharedDirectoryQueue.put(newDir);
	    }
	    catch (InterruptedException e){
		ExceptionHandler.handleExceptionExit(e);
	    }
	    Global.monitor.writeToBuffer("FileLister", "addDirectory", "exit");
	    
	}

    
    /**
     * Allows a peak at what is at the head of the queue
     * 
     * @return the file at the head of the queue, though not removed
     */

    public File peekFile(){
	File peek = this.fileQueue.peek();
	return peek;
    }
		  
    /**
     * Retrieves and removes the head of the file queue, waiting if necessary until an element becomes available
     * 
     * @return the file at the head of the file queue
     */
    public File takeFile() {	    
	    try {
		File take = this.fileQueue.take();
		//System.err.println("Thread " + Thread.currentThread().getName() + " taking file " + take);System.err.flush();
		return take;
	    } catch (InterruptedException e) {
		Global.monitor.writeToBuffer("FileLister", "takeFile", "thread_id = " + Thread.currentThread().getId());
		Global.monitor.writeToBuffer("FileLister", "takeFile", Arrays.toString(Thread.currentThread().getStackTrace()));
		ExceptionHandler.handleExceptionExit(e);
	    }
	    return null;
	}


    /**
     * Adds a file to the end of the queue as a marker for work being done
     * 
     * @param file
     *          the file to put at the tail of the file queue
     */
    public synchronized void putPoison(File poison) {
	    this.putFile(poison);
	}



    /**
     * Adds the file to the tail of the file queue, waiting if necessary for space to become available
     * 
     * @param file
     *          the file to put at the tail of the file queue
     */
    protected synchronized void putFile(File file) {
	    try {
		//System.err.println("Thread " + Thread.currentThread().getName() + " put file");System.err.flush();
		this.fileQueue.put(file);
	    } catch (InterruptedException e) {
		Global.monitor.writeToBuffer("FileLister", "putFile", "thread_id = " + Thread.currentThread().getId());
		Global.monitor.writeToBuffer("FileLister", "putFile", Arrays.toString(Thread.currentThread().getStackTrace()));
		ExceptionHandler.handleExceptionExit(e);
	    }
	}

    /**
     * Adds the file to the end of this list of files used for a dictionary
     * 
     * @param file
     *          the file to add to this list of files
     */
    protected synchronized void addFile(File file) {    

	    this.fileList.add(file);
	    Global.monitor.writeToBuffer("FileLister","addFile","File: " + file.toString());
	}
  
    /**
     * Returns a list of each of the files in this fileLister, each on a separate line
     */
    @Override
	public String toString() {
	StringBuffer string = new StringBuffer();
	for (File file : this.fileList) {
	    string.append(String.format("%s%n", file.toString()));
	}
	return string.toString();
    }

    /**
     * Recursively loads all the files from the given directory into the list of files used for a dictionary
     * 
     * @param path
     *          the string path to the directory to load
     */
    private void recursiveLoadDirectory(String path) {
	Global.monitor.writeToBuffer("FileLister", "recursiveLoadDirectory", "directory=" + path);    
    
	ArrayList<FileListerWorker> threads = new ArrayList<FileListerWorker>(maxThread);
	
	File poison = new File(new BigInteger(128, new Random()).toString());
	File root = new File(path);
	this.addDirectory(root);

	//Start some threads to read the directory structure looking for files
	// Make them daemon threads so that we don't have to figure out when they are done.
	for (int i = 0; i < maxThread; i++){
	    threads.add(i,new FileListerWorker(this,poison));
	    threads.get(i).setName("FileListerWorker " + i);
	    threads.get(i).setDaemon(true);
	    threads.get(i).start();
	}    

	FileListerHR  hr = new FileListerHR(threads,sharedDirectoryQueue,this,poison);
	hr.setDaemon(true);
	hr.setName("FileListerHR");
	hr.start();

	//        this.loaded = true;
	//	Global.monitor.printMessage("FileLister", "Files Loaded: " + root.getAbsolutePath());
	//Global.monitor.writeToBuffer("FileLister", "recursiveLoadDirectory", "loaded=" + this.loaded);
    }

    // }

    /**
     * Populates the dictionary and sample lists, waiting for all files to be loaded before randomly assigning files to
     * one of the lists. Dictionary and sample lists do not share any files. If this method is never called all files are
     * assigned to the dictionary list.
     * 
     * @param dictionarySize
     *          the <code>double</code> number of files to include in the dictionary list
     * @param sampleSize
     *          the <code>double</code> number of files to include in the sample list
     */
    synchronized public Pair<List<File>,List<File>> getSublists(double dictionarySize, double sampleSize) {

	while (!this.getLoaded()){
	    // Wait for the list to be loaded before copying it
	    try {
		// Not done yet		       
		Thread.sleep(100);						
	    } catch (InterruptedException e){
		ExceptionHandler.handleExceptionExit(e);
	    }
	}
	

	if (dictionarySize < 0 || sampleSize < 0) {
	    ExceptionHandler.handleExceptionThrow(new Exception("FileLister: Negative size parameter(s): dictionarySize = " + dictionarySize + "  sampleSize = " + sampleSize));
	}
	
	

	int dictionaryCount = (int) ((dictionarySize >= 1) ? dictionarySize
				     : ((dictionarySize * 10 * this.fileList.size()) + .5));
	int sampleCount = (int) ((sampleSize >= 1) ? sampleSize : ((sampleSize * 10 * this.fileList.size()) + .5));
	Global.monitor.printMessage("FileLister", "createSublists: dictionaryCount=" + dictionaryCount + "  sampleCount="
				    + sampleCount);
	Global.monitor.writeToBuffer("FileLister", "createSublists", "dictionaryCount=" + dictionaryCount);
	Global.monitor.writeToBuffer("FileLister", "createSublists", "sampleCount=" + sampleCount);

	if (dictionaryCount > this.fileList.size() - sampleCount
	    || sampleCount > this.fileList.size() - dictionaryCount) {
	    ExceptionHandler.handleExceptionThrow(new Exception(String.format("%s%n  %s%n  %s%n  %s%n",
									      "FileLister: Incompatible parameters: dictionary files + sample files <= total directory files",
									      "Total directory files: " + this.fileList.size(), "Requested number of dictionary files: "
									      + dictionaryCount, "Requested number of sample files: " + sampleCount)));
	}
	
	Collections.shuffle(this.fileList);

	ArrayList<File> outputDictionaryList = new ArrayList<File>();
	ArrayList<File> outputSampleList = new ArrayList<File>();
	outputDictionaryList.addAll(this.fileList.subList(0, dictionaryCount));
	outputSampleList.addAll(this.fileList.subList(dictionaryCount, dictionaryCount + sampleCount));
	Global.monitor.printMessage("FileLister", "createSublists: dictionaryList.size=" + this.dictionaryList.size()
				    + "  sampleList.size=" + this.sampleList.size());
	
	Global.monitor.printMessage("FileLister", "createSublists: completed");
	
	return(new Pair(outputDictionaryList, outputSampleList));

    }
}
